# dp_stm_map - Distributed and Persistent Software Transaction Map
# Copyright (C) 2013  Dragan Milic
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

require 'thread'
require 'set'

module DpStmMap

  class AtomicView

    def initialize global_state
      @global_state=global_state
      @local_state={}
      @to_delete=Set.new
    end

    def []= key, value
      if value == nil
        delete key
      else
        @local_state[key]=value
      end
    end

    def [] key
      if @local_state.has_key? key
        @local_state[key]
      else
        @local_state[key]=@global_state[key]
      end
    end

    def has_key? key
      if @local_state.has_key?(key)
        @local_state[key] != @global_state[key]
      else        
        (not @to_delete.include?(key)) && @global_state.has_key?(key)
      end
    end

    def delete key
      if @local_state.has_key? key
        @local_state.delete key
      else
        @to_delete << key
      end
    end

    def changes
      changes={}
      @local_state.each do |k,v|
        changes[k]=[@global_state[k],v]
      end
      @to_delete.each do |k|
        changes[k]=[@global_state[k],nil]
      end
      changes
    end

    def commit
      @to_delete.each do |key|
        @global_state.delete key
      end
      @global_state.merge!(@local_state)
    end

  end

  class AtomicReadView
    def initialize state
      @state=state
    end

    def [] key
      @state[key]
    end

    def has_key? key
      @state.has_key? key
    end
  end


  class InMemoryStmMap

    def initialize
      @mutex=Mutex.new
      @state={}
      @validators=[]
      @listeners=[]
    end


    def on_atomic &block
      @listeners << block
    end


    def validate_atomic &block
      @validators << block
    end


    def atomic timeout=nil
      @mutex.synchronize do
        view=AtomicView.new @state
        result=yield view

        changes=view.changes

        @validators.each do |validator|
          validator.call changes
        end

        view.commit

        @listeners.each do |listener|
          begin
            listener.call changes
          rescue
          end
        end

        result
      end
    end

    def atomic_read
      @mutex.synchronize do
        yield AtomicReadView.new @state
      end
    end


  end

end
